from .app import App
from .main import MainWindow
from .menu import MenuBar
from .progressbar import LoadingWindow
from .settings import SettingsWindow
from .about import AboutWindow
from .update import UpdateManager
from .error import run_error

__all__ = [
    "App",
    "MainWindow",
    "MenuBar",
    "LoadingWindow",
    "SettingsWindow",
    "AboutWindow",
    "UpdateManager",
    "run_error",
]


from PySide6 import QtWidgets, QtGui, QtCore
from numpy import True_
from .error import run_error

class SettingsWindow(QtWidgets.QDialog):
    """Window for settings."""
    def __init__(self, parent=None):
        super().__init__(parent)
        self.menu = parent
        self.setWindowTitle("Settings")
        self.setModal(True)
        
        # Settings Layout
        layout = QtWidgets.QVBoxLayout(self)
        layout.addLayout(self.init_color())
        layout.addLayout(self.init_highlight())
        layout.addLayout(self.init_max_sheets())
        layout.addLayout(self.init_date_format())
        layout.addLayout(self.init_oldest_to_newest())
        layout.addLayout(self.init_auto_update())
        layout.addSpacing(10)
        layout.addLayout(self.init_buttons())
        
    def init_color(self):
        """Initialize Hex color code for the background of the comments on Excel."""
        msg = "Hex color code for the background of the comments on Excel"
        regex = QtCore.QRegularExpression(r'^#[0-9A-Fa-f]{6}$')
        validator = QtGui.QRegularExpressionValidator(regex)
        
        # Color Label
        color_label = QtWidgets.QLabel("Background color :")
        color_label.setToolTip(msg)
        
        # Color1 input
        self.color1_input = QtWidgets.QLineEdit()
        self.color1_input.setText(self.menu.app.yt.settings.bg_color[0])
        self.color1_input.setPlaceholderText("#C5D9F1")
        self.color1_input.setToolTip(msg)
        self.color1_input.setValidator(validator)
        
        # Color2 input
        self.color2_input = QtWidgets.QLineEdit()
        self.color2_input.setText(self.menu.app.yt.settings.bg_color[1])
        self.color2_input.setPlaceholderText("#FDE9D9")
        self.color2_input.setToolTip(msg)
        self.color2_input.setValidator(validator)
        
        # Color Layout
        color_layout = QtWidgets.QHBoxLayout()
        color_layout.addWidget(color_label)
        color_layout.addWidget(self.color1_input)
        color_layout.addWidget(self.color2_input)
        return color_layout

    def init_highlight(self):
        """Initialize the toggle for the background color."""
        msg = "Check to activate the background color"
        
        # Highlight label
        highlight_label = QtWidgets.QLabel("Background color :")
        highlight_label.setToolTip(msg)
        
        # Highlight toggle
        self.highlight_toggle = QtWidgets.QCheckBox()
        self.highlight_toggle.setChecked(self.menu.app.yt.settings.bg_highlight)
        self.highlight_toggle.setToolTip(msg)
        
        # Highlight layout
        highlight_layout = QtWidgets.QHBoxLayout()
        highlight_layout.addWidget(highlight_label)
        highlight_layout.addWidget(self.highlight_toggle)
        return highlight_layout
    
    def init_max_sheets(self):
        """Initialize the maximum number of sheets per file to export."""
        msg = "Maximum number of sheets per file to export"
        
        # Sheets label
        sheets_label = QtWidgets.QLabel("Sheets per file :")
        sheets_label.setToolTip(msg)
        
        # Sheets input
        self.sheets_input = QtWidgets.QLineEdit()
        self.sheets_input.setText(str(self.menu.app.yt.settings.max_sheets))
        self.sheets_input.setToolTip(msg)
        
        # Sheets layout
        sheets_layout = QtWidgets.QHBoxLayout()
        sheets_layout.addWidget(sheets_label)
        sheets_layout.addWidget(self.sheets_input)
        return sheets_layout
    
    def init_date_format(self):
        """Initialize the date format."""
        msg = "Format of the date in the comments"
        
        # Date label
        date_label = QtWidgets.QLabel("Date format :")
        date_label.setToolTip(msg)
        
        # Date input
        self.date_input = QtWidgets.QLineEdit()
        self.date_input.setText(self.menu.app.yt.settings.date_format)
        self.date_input.setToolTip(msg)
        
        # Date layout
        date_layout = QtWidgets.QHBoxLayout()
        date_layout.addWidget(date_label)
        date_layout.addWidget(self.date_input)
        return date_layout
    
    def init_oldest_to_newest(self):
        """Initialize the toggle for the oldest to newest order."""
        msg = "Sort the comments and the videos"
        
        # Sort label
        sort_label = QtWidgets.QLabel("Sort order :")
        sort_label.setToolTip(msg)
        
        # Create dropdown menu
        self.sort_combo = QtWidgets.QComboBox()
        self.sort_combo.addItems(["Oldest to newest", "Newest to oldest"])
        self.sort_combo.setCurrentIndex(0 if self.menu.app.yt.settings.oldest_to_newest else 1)
        self.sort_combo.setToolTip(msg)
        
        # Sort layout
        sort_layout = QtWidgets.QHBoxLayout()
        sort_layout.addWidget(sort_label)
        sort_layout.addWidget(self.sort_combo)
        return sort_layout
    
    def init_auto_update(self):
        """Initialize the toggle for the automatic update."""
        msg = "Allow or not the program to search for software updates at startup"
        
        # Update label
        update_label = QtWidgets.QLabel("Search update at startup :")
        update_label.setToolTip(msg)
        
        # Update toggle
        self.update_toggle = QtWidgets.QCheckBox()
        self.update_toggle.setChecked(self.menu.app.yt.settings.auto_update)
        self.update_toggle.setToolTip(msg)
        
        # Update layout
        update_layout = QtWidgets.QHBoxLayout()
        update_layout.addWidget(update_label)
        update_layout.addWidget(self.update_toggle)
        return update_layout
    
    def init_buttons(self):
        """Initialize the OK and Cancel buttons."""
        # OK button
        ok_button = QtWidgets.QPushButton("OK")
        ok_button.clicked.connect(self.on_ok)
        
        # Cancel button
        cancel_button = QtWidgets.QPushButton("Cancel")
        cancel_button.clicked.connect(self.on_cancel)
        
        # Buttons layout
        buttons_layout = QtWidgets.QHBoxLayout()
        buttons_layout.addWidget(ok_button)
        buttons_layout.addWidget(cancel_button)
        return buttons_layout
    
    def on_ok(self):
        """Manage the click on the OK button."""
        
        # Validate the color inputs
        try:
            color1 = str(self.color1_input.text()).upper()
            color2 = str(self.color2_input.text()).upper()
            if color1 == "":
                color1 = "#C5D9F1"
                self.color1_input.setText("#C5D9F1")
            if color2 == "":
                color2 = "#FDE9D9"
                self.color2_input.setText("#FDE9D9")
                
            self.menu.app.yt.settings.bg_color = (color1, color2)
        except ValueError as e:
            run_error(e)
            self.color1_input.setText(str(self.menu.app.yt.settings.bg_color[0]))
            self.color2_input.setText(str(self.menu.app.yt.settings.bg_color[1]))
            return
        
        # Validate the max sheets inputs
        try:
            max_sheets = int(self.sheets_input.text())
            self.menu.app.yt.settings.max_sheets = max_sheets
        except ValueError as e:
            run_error("The maximum number of sheets must be at least 2")
            self.sheets_input.setText(str(self.menu.app.yt.settings.max_sheets))
            return
        
        # Validate the other settings
        self.menu.app.yt.settings.bg_highlight = bool(self.highlight_toggle.isChecked())
        self.menu.app.yt.settings.date_format = str(self.date_input.text())
        if self.sort_combo.currentText() == "Oldest to newest":
            self.menu.app.yt.settings.oldest_to_newest = True
        else:
            self.menu.app.yt.settings.oldest_to_newest = False
        self.menu.app.yt.settings.auto_update = bool(self.update_toggle.isChecked())
        self.menu.app.yt.settings.save()
        self.accept()
    
    def on_cancel(self):
        """Manage the click on the Cancel button."""
        self.reject()





import re
import ytComments.metadata as metadata
from PySide6 import QtWidgets, QtCore, QtGui
from .update import UpdateManager
from .error import run_error

class AboutWindow(QtWidgets.QDialog):
    """About menu of the application"""
    def __init__(self, parent):
        super().__init__(parent)
        self.menu = parent
        self.setWindowTitle("About")
        layout = QtWidgets.QVBoxLayout()

        self.init_version()
        layout.addLayout(self.layout_version)
        
        self.init_author()
        layout.addLayout(self.layout_author)
        
        self.init_license()
        layout.addLayout(self.layout_license)
        
        self.init_description()
        layout.addSpacing(10)
        layout.addLayout(self.layout_description)

        self.setLayout(layout)
        
    def init_version(self):
        """Version number and search update button."""
        self.layout_version = QtWidgets.QHBoxLayout()
        version_label = QtWidgets.QLabel("Version : ")
        version_text = QtWidgets.QLabel(metadata.version)

        # Add the button to check for updates
        self.check_update_button = QtWidgets.QPushButton("Check updates")
        self.check_update_button.setStyleSheet(self.style_link())
        self.check_update_button.clicked.connect(self.check_for_updates)
        
        self.layout_version.addWidget(version_label)
        self.layout_version.addWidget(version_text)
        self.layout_version.addWidget(self.check_update_button)
        self.layout_version.setAlignment(QtCore.Qt.AlignLeft)
        
    def init_author(self):
        """Program Author."""
        self.layout_author = QtWidgets.QHBoxLayout()
        author_label = QtWidgets.QLabel("Author : ")
        author_text = QtWidgets.QLabel(metadata.author)

        self.layout_author.addWidget(author_label)
        self.layout_author.addWidget(author_text)
        self.layout_author.setAlignment(QtCore.Qt.AlignLeft)

    def init_license(self):
        """License and code source url."""
        self.layout_license = QtWidgets.QHBoxLayout()
        license_label = QtWidgets.QLabel("License : ")
        license_text = QtWidgets.QLabel(metadata.license)
        
        url_button = QtWidgets.QPushButton("code source")
        url_button.setStyleSheet(self.style_link())
        url_button.clicked.connect(
            lambda: QtGui.QDesktopServices.openUrl(QtCore.QUrl(metadata.url))
            )

        self.layout_license.addWidget(license_label)
        self.layout_license.addWidget(license_text)
        self.layout_license.addWidget(url_button)
        self.layout_license.setAlignment(QtCore.Qt.AlignLeft)

    def init_description(self):
        """Software description."""
        self.layout_description = QtWidgets.QHBoxLayout()
        description_text = QtWidgets.QLabel(metadata.description)
        self.layout_description.addWidget(description_text)
        
    def style_link(self):
        """Style for links."""
        css = """
            QPushButton {
                border: none;  /* No border */
                background: none;  /* No background */
                color: DodgerBlue;  /* Text color */
                text-decoration: underline;  /* Underlined text to look like a link */
            }
            QPushButton:hover {
                color: RoyalBlue;  /* Text color on hover */
            }
        """
        return css

    def extract_repo_info(self, url: str):
        """Extracts the username and repository name from a GitHub URL."""
        pattern = r'https?://(?:www\.)?github\.com/([^/]+)/([^/]+)'
        match = re.match(pattern, url)

        if match:
            repo_owner = match.group(1)
            repo_name = match.group(2)
            return repo_owner, repo_name
        else:
            raise ValueError(
                "invalid URL"
                )

    def check_for_updates(self):
        """Check for update availability."""
        self.check_update_button.setEnabled(False)
        self.check_update_button.setText("Checking for updates...")
        QtWidgets.QApplication.processEvents()
        
        repo_owner, repo_name = self.extract_repo_info(metadata.url)
        updater = UpdateManager(repo_owner, repo_name, self)
        if updater.check_updates():
            self.check_update_button.setText("Downloading update...")
            QtWidgets.QApplication.processEvents()
            try:
                updater.update_software()
                self.check_update_button.setText("Update completed")
                updater.show_file_location_message(updater.new_filedir)
                self.menu.app.close()
            except Exception as e:
                msg = "An error occurred during the update"
                self.check_update_button.setText(msg)
                run_error(msg, details = e)
        else:
            self.check_update_button.setText("Software up to date")

        self.check_update_button.setEnabled(True)


import sys
from PySide6.QtWidgets import (
    QApplication, QDialog, QProgressBar, QLabel, QVBoxLayout, QPushButton
    )
from PySide6.QtCore import Qt

class LoadingWindow(QDialog):
    """Window to show the progress of a long-running operation."""
    def __init__(self, parent=None, countdown:bool = False):
        if QApplication.instance() is None:
                self.qapp = QApplication(sys.argv)
                
        super().__init__()
        self.setWindowTitle("Loading...")
        self.countdown = countdown
        
        if parent:
            self.parent = parent
            x = self.parent.geometry().x()
            y = self.parent.geometry().y()
        else:
            x = 100
            y = 100
        self.setGeometry(x, y, 300, 100)
        
        # Delete the Windows title buttons
        self.setWindowFlags(Qt.Window | Qt.WindowTitleHint | Qt.WindowStaysOnTopHint)

        # Initialize a close event
        self.closing = False

        # Create the progress bar
        self.progress_bar = QProgressBar(self)
        self.progress_bar.setMinimum(0)
        self.progress_bar.setMaximum(100)
        self.progress_bar.setStyleSheet("""
            QProgressBar {
                background-color: #f0f0f0;  /* Background color */
                border: 2px solid #8f8f8f;   /* Border color */
                border-radius: 5px;          /* Rounded corners */
                height: 15px;                 /* Height of the progress bar */
                text-align: center;          /* Center the text */
                color: #000000;              /* Text color */
            }
            QProgressBar::chunk {
                background-color: #3c8dbc;   /* Background color of the filled chunk */
                border-radius: 5px;          /* Rounded corners of the filled chunk */
            }
        """)
        
        # Add an explanatory text
        self.label = QLabel("Downloading...", self)
        self.label.setAlignment(Qt.AlignHCenter | Qt.AlignTop)
        
        # Add a countdown label
        if self.countdown:
            self.countdown_label = QLabel()
            self.countdown_label.setAlignment(Qt.AlignHCenter | Qt.AlignTop)

        # Add a cancel button
        msg = "Stop the download earlier, only the downloaded comments will be added to the old save"
        self.cancel_button = QPushButton("Cancel")
        self.cancel_button.clicked.connect(self.terminate)
        self.cancel_button.setToolTip(msg)

        # Create the layout
        layout = QVBoxLayout()
        layout.addWidget(self.progress_bar)
        layout.addWidget(self.label)
        if self.countdown: layout.addWidget(self.countdown_label)
        layout.addWidget(self.cancel_button)
        layout.setAlignment(Qt.AlignCenter)
        self.setLayout(layout)

    def update_progress(self, value, duration:int | None = None):
        """Update the progress bar value."""
        self.progress_bar.setValue(value)
        
        # Update the countdown label
        if self.countdown and duration is not None: 
            hours = int(duration // 3600)
            minutes = int((duration % 3600) // 60)
            seconds = int(duration % 60)
            
            self.countdown_label.setText(
                f"Remaining estimated time : {hours:02}h:{minutes:02}m:{seconds:02}s"
                )
        
        QApplication.processEvents()
        
    def setText(self, text):
        """Update the window title."""
        self.label.setText(text)
        
    def terminate(self):
        """Close the window."""
        self.closing = True
        self.close()
        
        
        


import json
from pathlib import Path
from screeninfo import get_monitors

class Settings():
    """Settings Manager to save and apply settings to the software."""
    def __init__(self):
        self._channel_url = 'https://www.youtube.com/@GabbPiano'
        self._directory = Path.home() / 'Desktop'
        self._bg_color = ('#C5D9F1', '#FDE9D9')
        self._bg_highlight = True
        self._max_sheets = 250
        self._date_format = 'DD/MM/YYYY'
        self._oldest_to_newest = True
        self._auto_update = True
        self._window_size = (100, 100, 320, 170)
    
    @property
    def channel_url(self):
        """URL of the YouTube channel."""
        return self._channel_url
    
    @channel_url.setter
    def channel_url(self, value):
        if value is None:
            self._channel_url = 'https://www.youtube.com/@GabbPiano'
            return
        elif not isinstance(value, str):
            raise TypeError(
                f"The channel URL must be a string, not {type(value)}"
                )
        self._channel_url = str(value)

    @property
    def directory(self):
        """Directory where the files will be saved."""
        return self._directory
    
    @directory.setter
    def directory(self, value):
        if value is None:
            return
        else:
            path = Path(value)
        
        if path.exists() and path.is_dir():
            self._directory = path
        else:
            raise ValueError(
                f"The directory {value} does not exist"
                )

    @property
    def bg_color(self):
        """Hex color code for the background of the comments on Excel."""
        return self._bg_color

    @bg_color.setter
    def bg_color(self, value):
        if (len(value) != 2
            or not isinstance(value, tuple)
            or not isinstance(value[0], str)
            or not isinstance(value[1], str)
            ):
            raise ValueError(
                "The value must be a tuple of two strings"
                )
        elif not (value[0].startswith('#') and len(value[0]) == 7):
            raise ValueError(
                "The first value must be a valid hex color code"
                )
        elif not (value[1].startswith('#') and len(value[1]) == 7):
            raise ValueError(
                "The second value must be a valid hex color code"
                )
        self._bg_color = value

    @property
    def bg_highlight(self):
        """Allow or not the background color to be highlighted."""
        return self._bg_highlight

    @bg_highlight.setter
    def bg_highlight(self, value):
        if not isinstance(value, bool):
            print("bg highlight value must be a boolean, default value will be used")
            return
        self._bg_highlight = value

    @property
    def max_sheets(self):
        """Maximum number of sheets per file to export."""
        return self._max_sheets

    @max_sheets.setter
    def max_sheets(self, value):
        if not isinstance(value, int):
            raise ValueError(
                "The maximum number of sheets must be an integer"
                )
        elif value < 2:
            raise ValueError(
                "The maximum number of sheets must be at least 2"
                )
        self._max_sheets = value

    @property
    def date_format(self):
        """Format of the date in the comments."""
        return self._date_format

    @date_format.setter
    def date_format(self, value):
        if not isinstance(value, str):
            raise ValueError(
                "the date format must be a string. E.g: 'DD/MM/YY' ; 'YYYY-MM-DD'"
                )
        self._date_format = value

    @property
    def oldest_to_newest(self):
        """Sort comments from oldest to newest."""
        return self._oldest_to_newest

    @oldest_to_newest.setter
    def oldest_to_newest(self, value):
        if not isinstance(value, bool):
            print("oldest to newest value must be a boolean, default value will be used")
            return
        self._oldest_to_newest = value

    @property
    def auto_update(self):
        """Allow or not the program to search for software updates at startup."""
        return self._auto_update

    @auto_update.setter
    def auto_update(self, value):
        if not isinstance(value, bool):
            print("auto update value must be a boolean, default value will be used")
            return
        self._auto_update = value

    @property
    def window_size(self):
        """Size of the window."""
        return self._window_size

    @window_size.setter
    def window_size(self, value):
        if not isinstance(value, tuple):
            return
        elif len(value) != 4:
            return
        
        # Check if the coordinates are within monitors coordinates
        monitors = get_monitors()
        x_min = 0
        y_min = 0
        x_max = 0
        y_max = 0
        for monitor in monitors:
            if monitor.x < x_min:
                x_min = monitor.x
            if monitor.y < y_min:
                y_min = monitor.y
            if monitor.x + monitor.width > x_max:
                x_max = monitor.x + monitor.width
            if monitor.y + monitor.height > y_max:
                y_max = monitor.y + monitor.height
        
        x_margin = 0
        y_margin = 0
        if value[0] < x_min + x_margin or value[0] + value[2] > x_max - x_margin:
            self._window_size = (100, 100, 320, 170)
            return
        if value[1] < y_min + y_margin or value[1] + value[3] > y_max - y_margin:
            self._window_size = (100, 100, 320, 170)
            return
        
        self._window_size = value

    def path_save(self):
        """Path to the saved parameters file."""
        file = "yt_parameters.json"
        folder = "Atem83"
        path_file = Path.home() / "Documents" / folder / file
        return path_file

    def save(self):
        """Save the parameters in a configuration file."""
        settings = {
            'channel_url': self.channel_url,
            'directory': str(self.directory),
            'bg_color': self.bg_color,
            'bg_highlight': self.bg_highlight,
            'max_sheets': self.max_sheets,
            'date_format': self.date_format,
            'oldest_to_newest': self.oldest_to_newest,
            'auto_update': self.auto_update,
            'window_size': self.window_size
            }
        path_file = self.path_save()

        try:
            # Create the folder if it doesn't exist
            path_folder = Path(path_file).parent
            path_folder.mkdir(parents=True, exist_ok=True)

            # Write my backup
            with open(path_file, 'w', encoding='utf-8') as file:
                json.dump(settings, file, indent=4, ensure_ascii=False)
        except:
            print("Save failed.")
            return

    def load(self):
        """Load the parameters from the configuration file."""
        path_file = self.path_save()
        try:
            with open(path_file, "r", encoding='utf-8') as file:
                settings = json.load(file)
                self.channel_url = settings["channel_url"]
                self.directory = settings["directory"]
                self.bg_color = tuple(settings["bg_color"])
                self.bg_highlight = settings["bg_highlight"]
                self.max_sheets = settings["max_sheets"]
                self.date_format = settings["date_format"]
                self.oldest_to_newest = settings["oldest_to_newest"]
                self.auto_update = settings["auto_update"]
                self.window_size = tuple(settings["window_size"])
        except:
            print("Load failed, the parameters by default will be used.")
            return



from PySide6.QtWidgets import QMenuBar, QMainWindow
from .about import AboutWindow
from .settings import SettingsWindow

class MenuBar(QMenuBar):
    """Menu bar of the application."""
    def __init__(self, parent: QMainWindow):
        super().__init__(parent)
        self.app = parent
        
        # Menu Settings
        self.settings_dialog = SettingsWindow(self)
        settings_menu = self.addAction("&Settings")
        settings_menu.triggered.connect(self.show_settings)
        
        # Menu About
        self.about_dialog = AboutWindow(self)
        help_menu = self.addAction("&About")
        help_menu.triggered.connect(self.about_dialog.exec)

    def show_settings(self):
        """Show the Settings window."""
        x = self.app.geometry().x() - 100
        y = self.app.geometry().y() - 100
        self.settings_dialog.setGeometry(x, y, 270, 200)
        self.settings_dialog.exec()


from configparser import ConfigParser
from pathlib import Path

def get_project_metadata():
    """Metadata grabber from setup.cfg"""
    config = ConfigParser()
    setup_cfg = Path(__file__).parent.parent / "setup.cfg"
    
    with open(setup_cfg, "r", encoding="utf-8") as f:
        config.read_file(f)
    
    return {
        "version": config.get("metadata", "version"),
        "author": config.get("metadata", "author"),
        "name": config.get("metadata", "name"),
        "description": config.get("metadata", "description"),
        "url": config.get("metadata", "url"),
        "license": config.get("metadata", "license"),
    }

# Initialize the globals variables for an easy import
globals().update(get_project_metadata())



from PySide6 import QtWidgets, QtCore
from .error import run_error
from ytComments.gui.progressbar import LoadingWindow

class MainWindow(QtWidgets.QFrame):
    """Main Window of the application."""
    def __init__(self, parent: QtWidgets.QMainWindow):
        super().__init__(parent)
        self.app = parent
        layout = QtWidgets.QVBoxLayout(self)
        
        # Initialize the layout
        layout.addLayout(self.init_url())
        layout.addLayout(self.init_old_save())
        layout.addLayout(self.init_new_save())
        layout.addLayout(self.init_exe())
        
    def init_url(self):
        """Initialize the URL."""
        url_label = QtWidgets.QLabel("URL :")
        self.url_input = QtWidgets.QLineEdit()
        self.url_input.setText(self.app.yt.settings.channel_url)
        layout = QtWidgets.QHBoxLayout()
        layout.addWidget(url_label)
        layout.addWidget(self.url_input)
        return layout

    def init_old_save(self):
        """Initialize the old save location."""
        old_save_label = QtWidgets.QLabel("Old save :")
        self.old_save_input = QtWidgets.QLineEdit()
        msg = "More than one save can be specified by separating them with a semicolon"
        self.old_save_input.setToolTip(msg)
        old_save_button = QtWidgets.QPushButton("Browse")
        old_save_button.clicked.connect(self.browse_file)
        layout = QtWidgets.QHBoxLayout()
        layout.addWidget(old_save_label)
        layout.addWidget(self.old_save_input)
        layout.addWidget(old_save_button)
        return layout
    
    def init_new_save(self):
        """Initialize the new save directory."""
        directory_label = QtWidgets.QLabel("Save Directory :")
        self.directory_input = QtWidgets.QLineEdit()
        self.directory_input.setText(str(self.app.yt.settings.directory))
        directory_button = QtWidgets.QPushButton("Browse")
        directory_button.clicked.connect(self.browse_directory)
        layout = QtWidgets.QHBoxLayout()
        layout.addWidget(directory_label)
        layout.addWidget(self.directory_input)
        layout.addWidget(directory_button)
        return layout
    
    def init_exe(self):
        """Initialize the execute button."""
        exe_button = QtWidgets.QPushButton("Execute")
        exe_button.clicked.connect(self.execute)
        layout = QtWidgets.QHBoxLayout()
        layout.addWidget(exe_button)
        return layout

    def browse_file(self):
        """Open a file dialog to select files."""
        files, _ = QtWidgets.QFileDialog.getOpenFileNames(
            self, "Select the old saves", "", "Excel Files (*.xlsx)")
        # Add them in the old save input
        if files:
            self.old_save_input.setText(";".join(files))
    
    def browse_directory(self):
        """Open a file dialog to select a directory."""
        directory = QtWidgets.QFileDialog.getExistingDirectory(
            self, "Select a directory"
            )
        # Add them in the directory input
        if directory:
            self.directory_input.setText(directory)
            self.app.yt.settings.directory = directory
            self.app.yt.settings.save()
            
    def execute(self):
        """Execute the program."""
        if self.url_input.text() == "":
            self.app.yt.settings.channel_url = None
        else:
            self.app.yt.settings.channel_url = self.url_input.text()
        old_saves = self.old_save_input.text().split(";")
        directory = self.directory_input.text()
        
        try:
            self.app.yt.refresh()
        except RuntimeError as e:
            self.app.yt.settings.save()
            run_error("The URL is invalid.\nInsert a valid URL.", details = e)
            return
        
        if self.app.yt.is_valid_youtube_channel() == False:
            self.app.yt.settings.save()
            run_error("The Channel URL is invalid.\nInsert a valid YouTube Channel URL.")
            return
        
        # Check the validity of the old saves
        waste = ""
        for save in old_saves.copy():
            file = QtCore.QFile(save)
            
            # If the file does not exist, remove it from the list
            if not file.exists():
                old_saves.remove(save)
                waste += save
                
                # Check if the waste is a valid file (eg: a file with ";" in its path)
                file = QtCore.QFile(waste)
                if file.exists():
                    old_saves.append(waste)
                    waste = ""
                else:
                    waste += ";"
                    file = QtCore.QFile(waste)
                    if file.exists():
                        old_saves.append(waste)
                        waste = ""
            
            # Check if the file is a .xlsx file
            if file.exists() and QtCore.QFileInfo(file).suffix() != "xlsx":
                run_error(
                    "Invalid file extension detected",
                    f"The file {file.fileName()} is not a .xlsx file"
                    )
                return
        
        # If there isn't any old save, set it to None
        if len(old_saves) == 0:
            old_saves = None
        else:
            self.app.yt.old_save = old_saves
        
        # Check if the directory exists
        dir = QtCore.QFile(directory)
        if not dir.exists():
            run_error(
                "Invalid directory",
                f"The directory {directory} does not exist"
                )
            return
        else:
            self.app.yt.settings.directory = directory
        
        # Save the settings
        self.app.yt.settings.save()
        
        # Reset somes values
        self.app.yt.finish = False
        self.app.yt.old_comments = None # Reset the previous old save
        
        # Show the loading window
        loading_window = LoadingWindow(self.app, countdown=True)
        loading_window.update_progress(0)
        loading_window.show()
        
        # Launch the program
        if old_saves is not None:
            try:
                self.app.yt.import_excel()
            except Exception as e:
                print(e)
                run_error("Invalid save, choose a valid save", details = str(e))
                loading_window.close()
                return
        self.app.yt.start() # Execute the thread
        
        while self.app.yt.finish == False:
            QtWidgets.QApplication.processEvents()
            loading_window.update_progress(
                self.app.yt.progress, 
                self.app.yt.duration
                )

            # Terminate the download earlier if the loading window is closed
            if loading_window.closing:
                self.app.yt.finish = True
        
        loading_window.close()
        
            
            


import yt_dlp
import polars as pl
import xlsxwriter
import time
from pathlib import Path
from colorama import Fore, init
from PySide6.QtCore import QThread, Signal
from .settingsManager import Settings

class yt_manager(QThread):
    """Download all comments from a YouTube channel,
    sorted by videos and export them in Excel files"""
    finished = Signal()
    
    def __init__(
        self, 
        channel_url: str | None = None, 
        dir: str | Path | None = None, 
        old_save: str | list[str] | None = None
        ):
        super().__init__()
        
        # Initialize the default values of parameters
        self.settings = Settings()
        self.settings.load()
        
        self.settings.channel_url = channel_url
        self.settings.directory = dir
        self._channel_data = None
            
        self.old_save = old_save
        self.old_comments = None
        self.progress = 0
        self.duration = None
        self.finish = False
        self._mono_video = None
            
    def __str__(self):
        init(autoreset=True)
        msg = f"{Fore.MAGENTA}Channel: {Fore.CYAN}{self.channel_title}\n"
        msg += f"{Fore.MAGENTA}Channel ID: {Fore.CYAN}{self.channel_id}\n"
        msg += f"\n{Fore.MAGENTA}{'Title':<80} {'Views':<10} {'URL'}\n"
        msg += f"{Fore.WHITE}-" * 140 + "\n"
        
        total_views = 0
        for video in self.channel_data['entries']:
            video_url = f"https://www.youtube.com/watch?v={video['id']}"
            video_title = video['title']
            video_views = video.get('view_count', 0)
            total_views += video_views
            msg += f"{Fore.YELLOW}{video_title[:80]:<80} {Fore.GREEN}{str(video_views)[:10]:<10} {Fore.CYAN}{video_url}\n"
        
        msg += f"\n{Fore.WHITE}{'-' * 140}\n"
        msg += f"{Fore.MAGENTA}Total Views for all videos: {Fore.GREEN}{total_views:,}"
        return msg
    
    def run(self):
        """Execution of the thread."""
        self.export_excel()
        self.finish = True
        self._mono_video = None # Reset the mono video
        self.finished.emit() # Emit finished signal
    
    @property
    def old_save(self) -> Path | list[Path]:
        """Returns the path of the old save."""
        return self._old_save
    
    @old_save.setter
    def old_save(self, value: str | Path | list[str | Path] | None):
        if value == None:
            return
        if isinstance(value, list):
            if all(isinstance(item, (str, Path)) for item in value):
                pass
            else:
                raise TypeError(
                    f"The list must only contain Path-like objects but is {type(value)}"
                    )
        elif isinstance(value, (str, Path)):
            value = [value]
        else:
            raise TypeError(
                f"The old save must be a Path-like object but is {type(value)}"
                )
        
        list_files = []
        for val in value:
            path = Path(val)
            if path.exists() and path.is_file():
                list_files.append(path)
            else:
                raise ValueError(
                    f"The file {val} does not exist"
                    )
        
        self._old_save = list_files
    
    @property
    def channel_data(self) -> pl.DataFrame:
        """Returns the data of the YouTube channel."""
        return self._channel_data
    
    @property
    def channel_title(self) -> str:
        """Returns the title of the YouTube channel."""
        channel_title = self.channel_data.get('title')
        if ' - Videos' in channel_title:
            channel_title = channel_title.replace(' - Videos', '')
        return channel_title

    @property
    def channel_id(self) -> str:
        """Returns the ID of the YouTube channel."""
        return self.channel_data.get('id')

    @property
    def channel_total_duration(self) -> str:
        """Returns the total duration of the channel videos in seconds."""
        total = self.channel_videos().select(pl.sum('duration')).to_series().item()

        # Convert total_seconds back to hh:mm:ss format if needed
        total_days = int(total // 86400)
        total_hours = int((total % 86400) // 3600)
        total_minutes = int((total % 3600) // 60)
        total_seconds = int(total % 60)
        
        formatted_time = f"{total_days:03}d:{total_hours:02}h:{total_minutes:02}m:{total_seconds:02}s"
        return formatted_time
    
    @property
    def channel_total_views(self) -> int:
        """Returns the total views of the channel videos."""
        return int(self.channel_videos().select('view_count').sum().item())
    
    @property
    def channel_number_videos(self) -> int:
        """Returns the number of videos of the channel."""
        return len(self.channel_videos())

    @property
    def videos_header(self) -> dict:
        """Returns the header row of the videos Excel sheet."""
        return {'row': 8, 'col': 0}

    @property
    def comments_header(self) -> dict:
        """Returns the header row of the comments Excel sheet."""
        return {'row': 6, 'col': 0}

    def refresh(self):
        """Refresh the channel data."""
        ydl_opts = {'quiet': True, 'extract_flat': True}
        try:
            with yt_dlp.YoutubeDL(ydl_opts) as ydl:
                self._channel_data = ydl.extract_info(self.settings.channel_url, download=False)
                
                # Allow to download only one video
                if self.channel_data.get('entries') is None:
                    self._mono_video = self._channel_data.get('id')
                    self.settings.channel_url = self._channel_data.get('channel_url')
                    self._channel_data = ydl.extract_info(self.settings.channel_url, download=False)
        except Exception as e:
            raise RuntimeError(
                f'Error fetching data:\n{e}'
                )

    def is_valid_youtube_channel(self) -> bool:
        """Check if the URL is a valid YouTube channel URL."""
        if (self.channel_data.get('_type') == 'playlist' and 
            self.channel_data.get('extractor').startswith('youtube')
            ):
            return True
        return False

    def channel_videos(self, include_comments: bool = False) -> pl.DataFrame:
        """Returns a pl.DataFrame with videos data of the channel."""
        
        df = pl.DataFrame(self.channel_data['entries'])
        df = df.select([
            'id', 
            'url', 
            'title', 
            'duration', 
            'view_count', 
            ])
        
        # from oldest to newest
        if self.settings.oldest_to_newest:
            df = df.reverse() 
        
        # Add a column with the comments
        total_videos = len(df)
        nb_video = 0
        self.progress = 0
        comments = []
        comments_count = []
        upload_dates = []
        start = time.time()
        
        if include_comments:
            for video in df.iter_rows():
                nb_video += 1
                
                # Download the new comments
                if ((self.finish == False and self._mono_video is None) or
                    (self._mono_video and self._mono_video == video[0])
                    ):
                    info = self.dl_comments(video)
                    comment = info['comments']
                    upload_date = info['upload_date']
                # if the download window is cancelled early, add the saved untouched comments
                elif (self.finish == True or self._mono_video) and self.old_comments:
                    comment = self.old_comments[0].get(video[0])
                    if comment is None: # if the video doesn't exist in the save, skip to the next loop
                        comments.append(None)
                        comments_count.append(0)
                        upload_dates.append(None)
                        continue
                    upload_date = self.old_comments[1].get(video[0])
                # if I'm in mono_video mode, continue the loop until the end
                elif self._mono_video:
                    comments.append(None)
                    comments_count.append(0)
                    upload_dates.append(None)
                    continue
                # if there isn't any save, finish the loop
                else:
                    break
                
                # Sort the comments
                comment = self._sort_comments(comment, self.settings.oldest_to_newest)
                comment = comment.with_columns(
                    pl.lit(self.settings.oldest_to_newest)
                    .alias('oldest_to_newest')
                    ) # save the sorting order to import later
                
                color_list = []
                current_color = 'color2'
                # Create a color column to switch color between each comment's thread
                # I failed to create it using polars expressions
                for value in comment['thread']:
                    if value == 'main':
                        if current_color == 'color1':
                            current_color = 'color2'
                        else:
                            current_color = 'color1'
                    color_list.append(current_color)

                comment = comment.with_columns(pl.Series('color', color_list))
                comments.append(comment)
                count = len(comment)
                comments_count.append(count)
                upload_dates.append(upload_date)
                
                # Update the progress bar
                self.progress = int(nb_video / total_videos * 100)
                self.duration = (time.time() - start) / nb_video * (total_videos - nb_video)
            
            # Add None values if not all videos are processed
            if len(comments) < total_videos:
                comments.extend([None] * (total_videos - len(comments)))

            # Add 0 values if not all videos are processed
            if len(comments_count) < total_videos:
                comments_count.extend([0] * (total_videos - len(comments_count)))

            # Add None values if not all videos are processed
            if len(upload_dates) < total_videos:
                upload_dates.extend([None] * (total_videos - len(upload_dates)))

            # Add the new columns
            df = df.with_columns(
                pl.Series(comments_count).alias('comments_count'),
                pl.Series(comments).alias('comments'),
                pl.Series(upload_dates).cast(pl.String).alias('upload_date')
                )
            
            # Convert upload_date from pl.String to pl.Date
            df = df.with_columns(
                pl.when(pl.col('upload_date').is_not_null())
                .then(pl.col('upload_date').str.strptime(pl.Date, format="%Y%m%d"))
                .otherwise(None)
                .alias('upload_date')
            )
        
        return df

    def dl_comments(self, video: tuple) -> pl.DataFrame:
        """Returns a pl.DataFrame with the comments of a YouTube video."""
        ydl_opts = {'getcomments': True}
        with yt_dlp.YoutubeDL(ydl_opts) as ydl:
            info = ydl.extract_info(video[1], download=False)
        
        upload_date = info.get('upload_date', None)
        
        # Create a dataframe from the comments
        df = pl.DataFrame(info['comments'])
        
        # Add ' when the text starts with = to avoid errors in Excel export
        df = df.with_columns(
            pl.when(pl.col('text').str.starts_with('='))
            .then("'" + pl.col('text'))
            .otherwise(pl.col('text'))
            .alias('text')
            )
        
        if len(df) != 0:
            # Add the columns date, id-parent, id-child
            df = df.with_columns(
                pl.from_epoch(pl.col('timestamp')).cast(pl.Date).alias('date'),
                pl.col('id')
                .str.split_exact('.', 1)
                .struct.rename_fields(['id-parent', 'id-child'])
                .alias('fields')
            ).unnest('fields')
            
            # Add the thread column
            df = df.with_columns(
                pl.when(pl.col('id-child').is_null())
                  .then(pl.lit('main'))
                  .otherwise(pl.lit('reply'))
                  .alias('thread')
            )
            
            df = df.select(['thread', 'author', 'date', 'text', 'id-parent', 'id-child'])
            # Add the new comments to the old comments
            if self.old_comments is not None:
                if video[0] in self.old_comments[0]:
                    comment = self.old_comments[0][video[0]]
                    matching_df = comment.join(
                        df, 
                        on=['id-parent', 'id-child'], 
                        how='inner', 
                        join_nulls=True, 
                        maintain_order='left'
                        )
                    unmatching_df = df.join(
                        matching_df, 
                        on=['id-parent', 'id-child'], 
                        how='anti', 
                        join_nulls=True, 
                        maintain_order='left'
                        )
                    df = pl.concat([unmatching_df, comment])
                
            df = df.select(['thread', 'author', 'date', 'text', 'id-parent', 'id-child'])
        else:
            # Create an empty dataframe if the video doesn't have any comments
            df = pl.DataFrame({
                'thread': [],
                'author': [],
                'date': [],
                'text': [],
                'id-parent': [],
                'id-child': [],
            })
        return {'comments': df, 'upload_date': str(upload_date)}

    def _sort_comments(self, df: pl.DataFrame, oldest_to_newest: bool) -> pl.DataFrame:
        """Sort the comments in the dataframe."""
        if not isinstance(df, pl.DataFrame):
            raise ValueError(
                "df must be a polars DataFrame"
                )
        if not isinstance(oldest_to_newest, bool):
            raise ValueError(
                "oldest_to_newest must be a boolean"
                )
        
        # Count the number of threads
        df = df.with_columns(
            pl.when(pl.col('thread') == 'main')
            .then(pl.col('thread').cum_count() + 1)
            .alias('order-main')
        )
        
        # Copy the number to their replies
        df = df.with_columns(
            pl.when(pl.col('thread') == 'reply')
            .then(pl.col('order-main').max().over('id-parent'))
            .otherwise(pl.col('order-main'))
        )
        
        # Count the number of comments
        df = df.with_row_index("order-reply", offset=2)
        
        # Rearrange the number when the reply (downloaded) is before the main (old)
        # => Give them the last number of the thread
        df = df.with_columns(
            pl.when((pl.col('thread') == 'reply') & (pl.col('order-reply') < pl.col('order-main')))
            .then(pl.col('order-reply').max().over('id-parent') + pl.col('order-reply').cum_sum().over('id-parent'))
            .otherwise(pl.col('order-reply'))
            .alias('order-reply')
        )
        
        # Assign the value 1 for the main of each thread
        df = df.with_columns(
            pl.when(pl.col('thread') == 'main')
            .then(1)
            .otherwise(pl.col('order-reply'))
            .alias('order-reply')
        )
        
        # Rank the comments in each thread
        df = df.with_columns(
            pl.when(pl.col('thread') == 'reply')
            .then(pl.col('order-reply').rank(method='ordinal').over('id-parent'))
            .otherwise(pl.col('order-reply'))
        )
        
        # Sort the comments by order
        df = df.sort(
            ['order-main', 'order-reply'], 
            descending=[oldest_to_newest, False],
            maintain_order=True
            ).drop(['order-main', 'order-reply'])
        
        return df

    def export_excel(self):
        """Export the channel data to an Excel file."""
        
        df = self.channel_videos(include_comments=True)
        save_name = 'Export ' + self.channel_title
        nb = self.settings.max_sheets - 1
        total_files = (len(df) + nb - 1) // nb # round up
        
        # Create the fitting number of Excel files
        for i in range(1, total_files + 1):
            start = (i - 1) * nb + 1
            if i == total_files:
                end = len(df)
            else:
                end = i * nb
            filtered_df = df.slice(start - 1, end - start + 1)
            
            path = Path(self.settings.directory / (save_name + f' {i}.xlsx'))
            with xlsxwriter.Workbook(path) as wb:
                hlink = {'align': 'left', 'font_color': 'blue', 'underline': 1}
                video_col = {
                    'id': {'align': 'left'},
                    'url': hlink,
                    'title': {'align': 'left'},
                    'upload_date': {'align': 'center', 'num_format': self.settings.date_format},
                    'duration': {'align': 'center'},
                    'view_count': {'align': 'center'},
                    'comments_count': {'align': 'center'}
                }
                comments_col = {
                    'thread': {'align': 'center'},
                    'child': {'align': 'center'},
                    'author': {'align': 'center'},
                    'date': {'align': 'center', 'num_format': self.settings.date_format},
                    'text': {'align': 'left', 'text_wrap': True}
                }
                video_header = self.videos_header
                comments_header = self.comments_header
                ws_video = wb.add_worksheet('videos')
                
                # Create some custom formats to use
                fmt_hyperlink = wb.add_format(hlink)
                fmt_left_align = wb.add_format({'align': 'left'})
                fmt_right_align = wb.add_format({'align': 'right'})
                fmt_center_align = wb.add_format({'align': 'center'})
                fmt_number_left = wb.add_format({'align': 'left', 'num_format': '#,##0'})
                fmt_number_center = wb.add_format({'align': 'center', 'num_format': '#,##0'})
                fmt_date_center = wb.add_format({'align': 'center', 'num_format': self.settings.date_format})
                fmt_channel_title = wb.add_format({
                    'align': 'center',
                    'valign': 'vcenter',
                    'bold': True,
                    'font_size': 20
                    })
                fmt_video_title = wb.add_format({
                    'align': 'left',
                    'bold': True,
                    'font_size': 14
                    })
                
                # Convert seconds to hh:mm:ss
                filtered_df = filtered_df.with_columns(
                    (pl.col('duration').cast(pl.Int64) // 3600).cast(pl.Utf8).str.zfill(2).alias('hh'),
                    (pl.col('duration').cast(pl.Int64) // 60 % 60).cast(pl.Utf8).str.zfill(2).alias('mm'),
                    (pl.col('duration').cast(pl.Int64) % 60).cast(pl.Utf8).str.zfill(2).alias('ss')
                )

                # Add a column with format hh:mm:ss
                filtered_df = filtered_df.with_columns(
                    (pl.concat_str([
                        pl.col('hh'), 
                        pl.lit('h:'), 
                        pl.col('mm'), 
                        pl.lit('m:'), 
                        pl.col('ss'),
                        pl.lit('s')
                        ])).alias('duration')
                )
                filtered_df = filtered_df.select([
                    'id', 
                    'url', 
                    'title',
                    'upload_date',
                    'duration', 
                    'view_count', 
                    'comments', 
                    'comments_count'
                ])
                
                # Export the videos summary
                filtered_df.drop(['comments']).write_excel(
                    workbook=wb,
                    worksheet=ws_video,
                    position=(video_header['row'], video_header['col']),
                    table_style='TableStyleMedium2',
                    column_formats=video_col,
                    header_format={'bold':True, 'align':'center'},
                    autofit=True,
                    freeze_panes=(video_header['row'] + 1, video_header['col']),
                    )
                
                # Add some custom values to the video sheet
                msg = f"Welcome to the YouTube channel of @{self.channel_title} !"
                ws_video.merge_range('A1:E1', 'Merged Cell')
                ws_video.write('A1', msg, fmt_channel_title)
                ws_video.write('B3', 'URL of the channel :', fmt_right_align)
                ws_video.write('B4', 'ID of the channel :', fmt_right_align)
                ws_video.write('B5', 'Total duration of the channel videos :', fmt_right_align)
                ws_video.write('B6', 'Total views of the channel :', fmt_right_align)
                ws_video.write('B7', 'Total number of videos of the channel :', fmt_right_align)
                ws_video.write('C3', self.settings.channel_url, fmt_hyperlink)
                ws_video.write('C4', self.channel_id, fmt_left_align)
                ws_video.write('C5', self.channel_total_duration, fmt_left_align)
                ws_video.write('C6', self.channel_total_views, fmt_number_left)
                ws_video.write('C7', self.channel_number_videos, fmt_left_align)
                
                filtered_df = filtered_df.with_row_index()
                # Export the comments of each video in a separate sheet
                for video in filtered_df.iter_rows():
                    comments = video[7]
                    if comments is not None:
                        ws_comment = wb.add_worksheet(video[1])
                        
                        # Limit the number of comments if they exceed Excel limit for a sheet
                        limit = 1048576 - self.comments_header['row'] - 1
                        if comments.shape[0] > limit:
                            comments = comments[:limit]
                         
                        # Add a hyperlink to each comments sheet in the summary
                        ws_video.write_url(
                            row=(video_header['row'] + int(video[0]) + 1), 
                            col=(video_header['col']), 
                            url=('internal:' + f"'{video[1]}'" + '!A1'), 
                            cell_format=fmt_hyperlink, 
                            string=video[1]
                            )
                        
                        row = comments_header['row'] + 2
                        col = chr(comments_header['col'] + 65 + comments.shape[1] - 1)
                        cell = f'${col}{row}'
                        formula1 = f'=IF({cell}="color1", TRUE, FALSE)'
                        formula2 = f'=IF({cell}="color2", TRUE, FALSE)'
                        
                        if self.settings.bg_highlight:
                            conditional_col = {
                                ('thread', 'author', 'date', 'text'): [{
                                    'type': 'formula',
                                    'criteria': formula1,
                                    'format': {'bg_color': self.settings.bg_color[0]}
                                },
                                {
                                    'type': 'formula',
                                    'criteria': formula2,
                                    'format': {'bg_color': self.settings.bg_color[1]}
                                }]
                            }
                        else:
                            conditional_col = None
                        
                        # Export the comments of the current video
                        comments.write_excel(
                            workbook=wb,
                            worksheet=ws_comment,
                            position=(comments_header['row'], comments_header['col']),
                            column_formats=comments_col,
                            conditional_formats=conditional_col,
                            header_format={'bold':True, 'align':'center'},
                            autofit=True,
                            hidden_columns=['color', 'oldest_to_newest', 'id-parent', 'id-child'],
                            freeze_panes=(comments_header['row'] + 1, comments_header['col'])
                        )
                                                
                        # Add some custom values to the comment sheet
                        ws_comment.write('A1', 'Title :', fmt_left_align)
                        ws_comment.write('A2', 'URL :', fmt_left_align)
                        ws_comment.write('A3', 'Upload :', fmt_left_align)
                        ws_comment.write('A4', 'Duration :', fmt_left_align)
                        ws_comment.write('A5', 'Views :', fmt_left_align)
                        ws_comment.write('B1', video[3], fmt_video_title)
                        ws_comment.write('B2', video[2])
                        ws_comment.write('B3', video[4], fmt_date_center)
                        ws_comment.write('B4', video[5], fmt_center_align)
                        ws_comment.write('B5', video[6], fmt_number_center)
                        ws_comment.write_url(
                            row=3, 
                            col=3, 
                            url='internal:videos!A1', 
                            cell_format=fmt_hyperlink, 
                            string='Back to the videos list'
                            )
    
    def import_excel(self):
        """Import old saved comments."""
        if self.old_save is None:
            raise ValueError(
                "An old_save value must be provided"
                )
        
        dict_data = {}
        upload_dict = {}
        for path in self.old_save:
            # Recover the upload date of each video
            upload_saved = pl.read_excel(
                path, 
                sheet_id=1, 
                read_options={"header_row": self.videos_header['row']}
                )
            upload_saved = upload_saved.with_columns(
                upload_saved["upload_date"].dt.strftime("%Y%m%d").alias("upload_date")
            )
            
            result_dict = upload_saved.select(['id', 'upload_date']).to_dict()
            for i in range(len(result_dict["id"])):
                upload_dict[result_dict["id"][i]] = result_dict["upload_date"][i]
            
            # Recover the comments of each video
            data_saved = pl.read_excel(
                path, 
                sheet_id=0, 
                read_options={"header_row": self.comments_header['row']}
                )
            
            if isinstance(data_saved, pl.DataFrame):
                raise TypeError(
                    f"The file {path} must contain more than one sheet"
                    )
            elif isinstance(data_saved, dict):
                data_saved.pop('videos', None)
                dict_data.update(data_saved)
        
        for key, df in dict_data.items():
            # Sort the comments to be 'newest to oldest' by default
            sort_order = df.get_column('oldest_to_newest')[0]
            if sort_order:
                df = self._sort_comments(df, sort_order)
                
            # Delete the useless columns to be recomputed and readded later
            df = df.drop('oldest_to_newest', 'color')
            
            # Delete some noises in the comments
            chars = ['_x000D_', '_x005F']
            for char in chars:
                df = df.with_columns(
                    df["text"].str.replace_all(char, '', literal=False)
                    )
            
            dict_data[key] = df
        
        self.old_comments = (dict_data, upload_dict)
        



import requests
import os
import sys
from PySide6.QtWidgets import QMessageBox
from packaging.version import Version
from .progressbar import LoadingWindow
import ytComments.metadata as metadata

class UpdateManager:
    """Update Downloader of the software"""
    def __init__(self, repo_owner, repo_name, parent=None):
        self.about = parent
        self.repo_owner = repo_owner
        self.repo_name = repo_name
        self._file_size = 0
        self._downloaded_size = 0
        self.dir_update = "update"

    @property
    def file_size(self):
        """Size of the file to download."""
        return self._file_size
    
    @file_size.setter
    def file_size(self, value:int):
        if not isinstance(value, int):
            msg = "The file size must be an integer"
            raise TypeError(
                msg
                )
        self._file_size = value

    @property
    def downloaded_size(self):
        """Size of the downloaded file."""
        return self._downloaded_size

    @downloaded_size.setter
    def downloaded_size(self, value:int):
        if not isinstance(value, int):
            msg = "The file size must be an integer"
            raise TypeError(
                msg
                )
        self._downloaded_size = value

    @property
    def progress(self):
        """Percentage of download."""
        if self.file_size == 0:
            return 0
        return int(self.downloaded_size / self.file_size * 100)

    def get_latest_release_info(self, extension=".exe"):
        """
        Retrieves the latest release information from GitHub, 
        including the tag and the URL of the .exe file.

        Returns:
            tuple: A tuple containing the latest version tag and 
            the URL of the .exe file, or (None, None) if an error occurs.
        """
        url = f"https://api.github.com/repos/{self.repo_owner}/{self.repo_name}/releases/latest"
        try:
            response = requests.get(url)
            response.raise_for_status()  # Throws an exception for HTTP error codes
            data = response.json()
            tag_name = data.get("tag_name")
            assets = data.get("assets", [])
            exe_url = None
            for asset in assets:
                if asset.get("name", "").lower().endswith(extension):
                    exe_url = asset.get("browser_download_url")
                    break
            return tag_name, exe_url
        except requests.exceptions.RequestException as e:
            print(f"Error while retrieving the latest version from GitHub : {e}")
            return None, None

    def check_updates(self):
        """
        Checks if an update is available.

        Returns:
            bool: True if an update is available, False otherwise.
        """
        latest_tag, exe_url = self.get_latest_release_info()
        
        if not latest_tag or not exe_url:
            return False
        
        latest_version = Version(latest_tag.lstrip('v'))
        current_version = Version(metadata.version)

        return latest_version > current_version

    def update_software(self):
        """Updates the software by downloading the .exe file from GitHub."""
        
        print("Updating software...")
        _, exe_url = self.get_latest_release_info()
        if not exe_url:
            print("No .exe file founded in the latest version.")
            return
        print(f"Downloading the new version from : {exe_url}")
        
        response = requests.get(exe_url, stream=True)
        response.raise_for_status()
        self.file_size = int(response.headers.get("content-length", 0))
        
        # Check if the program is an executable or a python file
        if hasattr(sys, 'frozen'):
            self.old_path = os.path.abspath(sys.executable) # .exe
        else:
            self.old_path = os.path.abspath(__file__) # .py
        
        # Define the directory and file path for the update
        new_filename = str(os.path.basename(exe_url))
        self.new_filedir = os.path.join(
            os.path.dirname(self.old_path), 
            self.dir_update
            )
        self.new_path = os.path.join(self.new_filedir, new_filename)
        
        # Create the directory if it doesn't exist
        os.makedirs(self.new_filedir, exist_ok=True)
        
        # Display the progress bar
        loading_window = LoadingWindow(self.about)
        loading_window.show()
        
        # Download the updated file
        with open(self.new_path, "wb") as f:
            for chunk in response.iter_content(chunk_size=8192):
                f.write(chunk)
                self.downloaded_size += len(chunk)
                loading_window.update_progress(self.progress)
                print(f"Progress update : {self.progress}%", end="\r")
        
        loading_window.close()
        print("Download of the update completed.")
    
    def show_file_location_message(self, file_path):
        """Shows a message to the user about the new version."""
        msg_box = QMessageBox()
        txt = f"""
        The update has been downloaded here :\n\n
        {file_path}\n
        You can delete the old version.
        """
        msg_box.setWindowTitle("Update completed")
        msg_box.setText(txt)
        msg_box.setIcon(QMessageBox.Information)
        msg_box.setStandardButtons(QMessageBox.Ok)

        # Display the message
        msg_box.exec()




import sys
from PySide6.QtWidgets import (QApplication, QDialog, QLabel, QPushButton,
                               QVBoxLayout, QHBoxLayout, QSizePolicy, QWidget,
                               QMessageBox)
from PySide6.QtCore import Qt

class ErrorWindow(QDialog):
    def __init__(self, error_message:str = "Une erreur est survenue", error_details:str = ""):
        super().__init__()

        self.setWindowTitle("Erreur")
        self.setModal(True)
        self.layout = QVBoxLayout()
        
        # Définir le drapeau pour rester au premier plan
        self.setWindowFlags(self.windowFlags() | Qt.WindowStaysOnTopHint)
        
        # Icone d'erreur
        warning_icon = QMessageBox().standardIcon(QMessageBox.Warning)
        icon_label = QLabel()
        icon_label.setPixmap(warning_icon)
        icon_label.setAlignment(Qt.AlignLeft)
        
        message_layout = QHBoxLayout()
        message_layout.addWidget(icon_label)

        # Message d'erreur simplifié
        error_label = QLabel(error_message)
        error_label.setAlignment(Qt.AlignLeft)
        message_layout.addWidget(error_label)
        self.layout.addLayout(message_layout)

        # Bouton pour afficher le message d'erreur
        self.toggle_button = QPushButton("Afficher le message d'erreur")
        self.toggle_button.clicked.connect(self.toggle_message)
        if error_details != "":
            self.layout.addWidget(self.toggle_button)

        # Message d'erreur détaillé
        self.error_details_label = QLabel(error_details)
        self.error_details_label.setVisible(False)
        self.layout.addWidget(self.error_details_label)

        # Permet de dimensionner en petit le bouton "OK"
        button_layout = QHBoxLayout()
        spacer = QWidget()
        spacer.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)
        button_layout.addWidget(spacer)

        # Bouton de fermeture
        close_button = QPushButton("OK")
        close_button.clicked.connect(self.close)
        button_layout.addWidget(close_button)
        self.layout.addLayout(button_layout)

        self.setLayout(self.layout)
        self.adjustSize()

        # True = fenêtre détaillée, False = fenêtre simplifiée
        self.expanded = False  

    def toggle_message(self):
        """Affiche ou masque le message d'erreur."""
        
        if self.expanded:
            self.toggle_button.setText("Afficher le message d'erreur")
            self.error_details_label.setVisible(False)
            self.adjustSize()
        else:
            self.toggle_button.setText("Masquer le message d'erreur")
            self.error_details_label.setVisible(True)
            self.adjustSize()
        self.expanded = not self.expanded

def run_error(message:str = "", details:str = ""):
    """Lance la fenêtre d'erreur."""
    app = QApplication.instance()
    # Créez une nouvelle instance si nécessaire
    if app is None:
        app = QApplication(sys.argv)

    window = ErrorWindow(str(message), str(details))
    window.exec()


import sys
import ytComments.metadata as metadata
from PySide6.QtWidgets import QApplication, QMainWindow, QMessageBox
from ytComments import yt_manager
from .menu import MenuBar
from .main import MainWindow
from .update import UpdateManager
from .error import run_error

class App(QMainWindow):
    """Main application class."""
    def __init__(self):
        # Initialize QApplication
        self.qapp = QApplication(sys.argv)
        self.yt = yt_manager()
        self.yt.settings.load()
        
        # Initialize QMainWindow
        super().__init__()
        self.setWindowTitle(metadata.name)
        self.setGeometry(
            self.yt.settings.window_size[0],
            self.yt.settings.window_size[1],
            self.yt.settings.window_size[2],
            self.yt.settings.window_size[3]
        )
        
        # Create the main window
        self.main_frame = MainWindow(self)
        self.setCentralWidget(self.main_frame)
        
        # Create the menu bar
        self.menu_bar = MenuBar(self)
        self.setMenuBar(self.menu_bar)
        
        # Connect the window resize event and move event
        self.resizeEvent = self.on_resize
        self.moveEvent = self.on_move
        
        # Allow the window to accept drag-and-drop events for files
        self.setAcceptDrops(True)
        
    def run(self):
        """Execution of the application."""
        self.show()
        self.check_for_updates()
        sys.exit(self.qapp.exec())
        
    def on_resize(self, event):
        """Execute actions when the main window is resized."""
        self.yt.settings.window_size = (
            self.x(),
            self.y(),
            self.width(), 
            self.height()
            )
        self.yt.settings.save()
        super().resizeEvent(event)
        
    def on_move(self, event):
        """Execute actions when the main window is moved."""
        self.yt.settings.window_size = (
            self.x(),
            self.y(),
            self.width(), 
            self.height()
            )
        self.yt.settings.save()
        super().moveEvent(event)
        
    def dragEnterEvent(self, event):
        """Manage the drag-and-drop event.
        Only accept files."""
        if event.mimeData().hasUrls():
            event.accept()  # Accept the event
        else:
            event.ignore()  # Ignore if it's not a file
    
    def dropEvent(self, event):
        """Manage the drop event and retrieve the dropped file."""
        if event.mimeData().hasUrls():
            # Retrieve the URLs from the mimeData
            urls = event.mimeData().urls()
            # Convert the URLs to a local path
            list_urls = []
            for url in urls:
                file_path = url.toLocalFile()
                list_urls.append(file_path)
            # Set the path in the QLineEdit
            self.main_frame.old_save_input.setText(";".join(list_urls))

    def check_for_updates(self):
        """Check for updates."""
        repo_owner, repo_name = self.menu_bar.about_dialog.extract_repo_info(
            metadata.url
            )
        updater = UpdateManager(repo_owner, repo_name, self)
        reply = None
        
        # Check for updates
        if updater.check_updates() and self.yt.settings.auto_update:
            msg = "An update is available,"
            msg += "\nwould you like to update the software ?"
            reply = QMessageBox.question(
                None, 
                "Information", 
                msg, 
                QMessageBox.Yes | QMessageBox.No
                )
        
        # Download the update
        if updater.check_updates() and reply == QMessageBox.Yes:
            try:
                updater.update_software()
                updater.show_file_location_message(updater.new_filedir)
                self.close()
                sys.exit()
            except Exception as e:
                msg = "An error occurred during the update"
                run_error(msg, details = e)
                
